"""Test the functions related to the internal interface implementation and the 'Interface' interface itself
"""

from __future__ import annotations

from logging import getLogger
from pathlib import Path
from typing import Any

import pytest
from cppython_core.exceptions import PluginError
from cppython_core.schema import (
    CoreData,
    CPPythonLocalConfiguration,
    PEP621Configuration,
    ProjectConfiguration,
    PyProject,
)
from pytest_cppython.fixtures import CPPythonFixtures
from pytest_cppython.mock.generator import MockGenerator
from pytest_cppython.mock.provider import MockProvider
from pytest_mock import MockerFixture

from cppython.builder import Builder
from cppython.project import Project


class TestProject(CPPythonFixtures):
    """Grouping for Project class testing"""

    def test_construction_without_plugins(
        self, mocker: MockerFixture, project: PyProject, project_configuration: ProjectConfiguration
    ) -> None:
        """Verification that no error is thrown and output is gracefully handled if no provider plugins are found

        Args:
            mocker: Mocking fixture for interface mocking
            project: PyProject data to construct with
            project_configuration: Temporary workspace for path resolution
        """

        interface_mock = mocker.MagicMock()
        with pytest.raises(PluginError):
            Project(project_configuration, interface_mock, project.dict(by_alias=True))

    def test_construction_with_plugins(
        self, mocker: MockerFixture, project_configuration: ProjectConfiguration, project_with_mocks: dict[str, Any]
    ) -> None:
        """Verification of full construction with mock provider plugin

        Args:
            mocker: Mocking fixture for interface mocking
            project_configuration: Temporary workspace for path resolution
            project_with_mocks: PyProject data to construct with
        """

        mocked_provider_list = [MockProvider]
        mocker.patch("cppython.builder.Builder.discover_providers", return_value=mocked_provider_list)

        mocked_generator_list = [MockGenerator]
        mocker.patch("cppython.builder.Builder.discover_generators", return_value=mocked_generator_list)

        interface_mock = mocker.MagicMock()
        project_configuration.version = None
        Project(project_configuration, interface_mock, project_with_mocks)


class TestBuilder(CPPythonFixtures):
    """Tests of builder steps"""

    def test_plugin_gather(self) -> None:
        """Verifies that discovery works with no results"""

        builder = Builder(getLogger())
        providers = builder.discover_providers()

        assert len(providers) == 0

        generators = builder.discover_generators()

        assert len(generators) == 0

        vcs = builder.discover_vcs()

        assert len(vcs) == 1

    def test_provider_creation(
        self,
        core_data: CoreData,
        project_with_mocks: dict[str, Any],
    ) -> None:
        """Test that providers can be created with the mock data available

        Args:
            core_data: The resolved configuration data
            project_with_mocks: Data with mocks attached
        """

        builder = Builder(getLogger())

        provider_configurations = project_with_mocks["tool"]["cppython"]["provider"]

        providers = builder.create_providers(
            [MockProvider],
            core_data,
            provider_configurations,
        )

        assert len(providers) == 1

    def test_generator_creation(
        self,
        core_data: CoreData,
        project_with_mocks: dict[str, Any],
    ) -> None:
        """Test that providers can be created with the mock data available

        Args:
            core_data: The resolved configuration data
            project_with_mocks: Local config
        """

        builder = Builder(getLogger())

        generator_configurations = project_with_mocks["tool"]["cppython"]["generator"]

        assert builder.create_generator(
            [MockGenerator],
            core_data,
            generator_configurations,
        )

    def test_core_data_version(self) -> None:
        """Test the VCS config error override. Validated data is already tested."""

        builder = Builder(getLogger())

        project_configuration = ProjectConfiguration(pyproject_file=Path("pyproject.toml"), version=None)
        pep621_configuration = PEP621Configuration(name="version-resolve-test", dynamic=["version"], version=None)
        cppython_configuration = CPPythonLocalConfiguration()

        core_data = builder.generate_core_data(project_configuration, pep621_configuration, cppython_configuration)

        assert core_data.pep621_data.version



"""Manages data flow to and from plugins
"""

import asyncio
import logging
from typing import Any

from cppython_core.exceptions import ConfigError, PluginError
from cppython_core.plugin_schema.interface import Interface
from cppython_core.schema import CoreData, ProjectConfiguration, PyProject

from cppython.builder import Builder
from cppython.schema import API


class Project(API):
    """The object constructed at each entry_point"""

    def __init__(
        self, configuration: ProjectConfiguration, interface: Interface, pyproject_data: dict[str, Any]
    ) -> None:
        self._enabled = False
        self._interface = interface

        # Default logging levels
        levels = [logging.WARNING, logging.INFO, logging.DEBUG]

        # Add default output stream
        self.logger = logging.getLogger("cppython")
        self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(levels[configuration.verbosity])

        self.logger.info("Initializing project")

        if (pyproject := PyProject(**pyproject_data)) is None:
            raise ConfigError("PyProject data is not defined")

        builder = Builder(self.logger)

        if not (provider_plugins := builder.discover_providers()):
            raise PluginError("No provider plugin was found")

        for provider_plugin in provider_plugins:
            self.logger.warning("Provider plugin found: %s", provider_plugin.name())

        if not (generator_plugins := builder.discover_generators()):
            raise PluginError("No generator plugin was found")

        for generator_plugin in generator_plugins:
            self.logger.warning("Generator plugin found: %s", generator_plugin.name())

        if pyproject.tool is None:
            raise ConfigError("Table [tool] is not defined")

        if pyproject.tool.cppython is None:
            raise ConfigError("Table [tool.cppython] is not defined")

        self._core_data = builder.generate_core_data(configuration, pyproject.project, pyproject.tool.cppython)

        self._providers = builder.create_providers(provider_plugins, self.core_data, pyproject.tool.cppython.provider)
        self._generator = builder.create_generator(generator_plugins, self.core_data, pyproject.tool.cppython.generator)

        for provider in self._providers:
            assert provider.supports_generator(self._generator.name())

        self._enabled = True

        self.logger.info("Initialized project successfully")

    @property
    def enabled(self) -> bool:
        """Queries if the project was is initialized for full functionality

        Returns:
            The query result
        """
        return self._enabled

    @property
    def core_data(self) -> CoreData:
        """Queries if the project was is initialized for full functionality

        Returns:
            The query result
        """
        return self._core_data

    async def download_provider_tools(self) -> None:
        """Download the provider tooling if required"""
        if not self._enabled:
            self.logger.info("Skipping 'download_provider_tools' because the project is not enabled")
            return

        base_path = self.core_data.cppython_data.install_path

        for provider in self._providers:
            path = base_path / provider.name()

            path.mkdir(parents=True, exist_ok=True)

            if not provider.tooling_downloaded(path):
                self.logger.warning("Downloading the %s requirements to %s", provider.name(), path)

                await provider.download_tooling(path)
                self.logger.warning("Download complete")
            else:
                self.logger.info("The %s provider is already downloaded", provider.name())

    def sync(self) -> None:
        """Gathers sync information from providers and passes it to the generator"""

        inputs = []
        for provider in self._providers:
            inputs.append(provider.sync_data(self._generator.name()))

        self._generator.sync(inputs)

    # API Contract
    def install(self) -> None:
        """Installs project dependencies

        Raises:
            Exception: Raised if failed
        """
        if not self._enabled:
            self.logger.info("Skipping install because the project is not enabled")
            return

        self.logger.info("Installing tools")
        asyncio.run(self.download_provider_tools())

        self.logger.info("Installing project")

        for provider in self._providers:
            self.logger.info("Installing %s provider", provider.name())

            try:
                provider.install()
            except Exception as exception:
                self.logger.error("Provider %s failed to install", provider.name())
                raise exception

        self.sync()

    def update(self) -> None:
        """Updates project dependencies

        Raises:
            Exception: Raised if failed
        """
        if not self._enabled:
            self.logger.info("Skipping update because the project is not enabled")
            return

        self.logger.info("Updating tools")
        asyncio.run(self.download_provider_tools())

        self.logger.info("Updating project")

        for provider in self._providers:
            self.logger.info("Updating %s provider", provider.name())

            try:
                provider.update()
            except Exception as exception:
                self.logger.error("Provider %s failed to update", provider.name())
                raise exception

        self.sync()
